// SPDX-License-Identifier: GPL-2.0-or-later
/*
<URL:http://gcc.gnu.org/ml/gcc/1999-04n/msg01105.html>

Re: changing embedded RPATH in existing executables.

  To: geoffk@ozemail.com.au
  Subject: Re: changing embedded RPATH in existing executables.
  From: <peeter_joot@VNET.IBM.COM> (peeter joot)
  Date: Fri, 30 Apr 1999 16:14:44 -0400 (EDT)
  Cc: peeterj@ca.ibm.com, egcs@cygnus.com, libc-hacker@cygnus.com, linux-gcc@vger.rutgers.edu
  Reply-To: <peeter_joot@VNET.IBM.COM>

> _Changing_ is a little tricky, but the attached program strips rpaths
> from executables (I find it essential for debugging the binutils).
> It's endian-dependent, if you want this for x86 you can just change
> the occurrences of 'MSB' to 'LSB' and compile (I should really fix
> that).

Hi Geoff,

With your program as a guide (and some peeks into libbfd, elf.h, a bit
of the glibc dynamic loader code, objdump, and a hex-editor) I was able to
figure out enough to find and change the rpath string.  That was fun!

This program assumes (unlike your original program) that there is only
one DT_RPATH tag in the dynamic section as even with multiple '-Wl,-rpath,'
commands in the link this seems to occur (they all get concatonated into
a : separated path).

Thanks for your help.  If you want to use this on non-x86 you have to change
the occurances of LSB back to MSB:)

Peeter
--
*/

#include "config.h"
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <elf.h>
#if defined(HAVE_LINK_H)
#  include <link.h>
#endif /* HAVE_LINK_H */
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include "protos.h"

/**
 * Reads an ELF file, and reads or alters the RPATH setting.
 *
 * TODO:
 *  modify to add RPATH setting if none exists.
 */


/**
 * Reads the section names table from an ELF file into memory
 */
static char *
read_section_names(int fd, ChrElf_Ehdr ehdr)
{
  ChrElf_Shdr shdr;

  const size_t sz_shdr = is_e32 ? sizeof(Elf32_Shdr) : sizeof(Elf64_Shdr);
  const size_t sh_off = EHDRWU(e_shoff);
  const size_t sections_off = EHDRWU(e_shstrndx) * sz_shdr;

  if (pread(fd, &shdr, sz_shdr, sh_off + sections_off) != (ssize_t)sz_shdr)
  {
    perror ("reading section header");
    return NULL;
  }

  /* +1 for forced trailing null */
  char* strtab = malloc(SHDR_O(sh_size)+1);
  if (strtab == NULL)
    {
      perror ("allocating memory for string table");
      return NULL;
    }

  if (pread(fd, strtab, SHDR_O(sh_size), SHDR_O(sh_offset)) != (ssize_t)SHDR_O(sh_size))
  {
    perror ("reading string table");
    return NULL;
  }
  strtab[SHDR_O(sh_size)] = 0; /* make sure printed string is null terminated */
  return strtab;
}

/**
 * Looks up the name for a section, using its offset in the table
 */
static char *
get_section_name(size_t name_off, char *section_names) {
  return section_names + name_off;
}

int
chrpath(const char *filename, const char *newpath, bool convert)
{
  int fd;
  ChrElf_Ehdr ehdr;
  int i, ret = 0;
  ChrElf_Phdr phdr;
  ChrElf_Shdr shdr;
  void *dyns = NULL;
  int rpathoff;
  char * strtab = NULL;
  char * section_names = NULL;
  char * rpath;
  unsigned int rpathlen;
  int oflags;
  int rpath_dyns_index;

  if (!newpath && !convert)
     oflags = O_RDONLY;
  else
     oflags = O_RDWR;

  fd = elf_open(filename, oflags, &ehdr);
  if (fd == -1)
  {
    perror ("elf_open");
    return 1;
  }

   if (0 != elf_find_dynamic_section(fd, &ehdr, &phdr))
   {
     perror("found no dynamic section"), ret = 1;
     goto ret;
   }

  dyns = malloc(PHDR(p_filesz));
  if (dyns == NULL)
    {
      perror ("allocating memory for dynamic section"), ret = 1;
      goto ret;
    }
  if (pread(fd, dyns, PHDR(p_filesz), PHDR(p_offset)) != (ssize_t)PHDR(p_filesz))
    {
      perror ("reading dynamic section"), ret = 1;
      goto ret;
    }

  rpathoff = -1;
  for ( rpath_dyns_index = 0; DYNSS(rpath_dyns_index, d_tag) != DT_NULL;
        ++rpath_dyns_index )
    {
      if ( elf_dynpath_tag(DYNSS(rpath_dyns_index, d_tag)) )
      {
         rpathoff = DYNSU(rpath_dyns_index, d_un.d_ptr);
         break;
      }
    }
  if (rpathoff == -1)
    {
      printf("%s: no rpath or runpath tag found.\n", filename), ret = 2;
      goto ret;
    }

  section_names = read_section_names(fd, ehdr);
  if (section_names == NULL) {
    ret = 1;
    goto ret;
  }

  if (lseek(fd, EHDRWU(e_shoff), SEEK_SET) == -1)
  {
    perror ("positioning for sections"), ret = 1;
    goto ret;
  }

  for (i = 0; i < EHDRHU(e_shnum); i++)
  {
    const size_t sz_shdr = is_e32 ? sizeof(Elf32_Shdr) : sizeof(Elf64_Shdr);
    if (read(fd, &shdr, sz_shdr) != (ssize_t)sz_shdr)
    {
      perror ("reading section header"), ret = 1;
      goto ret;
    }
    if (SHDR_W(sh_type) == SHT_STRTAB
        && strcmp(get_section_name(SHDR_W(sh_name), section_names), ".dynstr") == 0)
      break;
  }
  if (i == EHDRHU(e_shnum))
    {
      fprintf (stderr, "No string table found.\n"), ret = 2;
      goto ret;
    }
  /* +1 for forced trailing null */
  strtab = malloc(SHDR_O(sh_size)+1);
  if (strtab == NULL)
    {
      perror ("allocating memory for string table"), ret = 1;
      goto ret;
    }

  if (pread(fd, strtab, SHDR_O(sh_size), SHDR_O(sh_offset)) != (ssize_t)SHDR_O(sh_size))
  {
    perror ("reading string table"), ret = 1;
    goto ret;
  }
  strtab[SHDR_O(sh_size)] = 0; /* make sure printed string is null terminated */

  if ((int)SHDR_O(sh_size) < rpathoff)
  {
    fprintf(stderr, "%s string offset not contained in string table\n",
            elf_tagname(DYNSS(rpath_dyns_index, d_tag))),
      ret = 5;
    goto ret;
  }
  rpath = strtab+rpathoff;

#if defined(DT_RUNPATH)
  if (convert && DYNSS(rpath_dyns_index, d_tag) == DT_RPATH)
  {
    if (is_e32)
      ((Elf32_Dyn *)dyns)[rpath_dyns_index].d_tag = swap_bytes ?
        __builtin_bswap32(DT_RUNPATH) : DT_RUNPATH;
    else
      ((Elf64_Dyn *)dyns)[rpath_dyns_index].d_tag = swap_bytes ?
        __builtin_bswap64(DT_RUNPATH) : DT_RUNPATH;
    if (pwrite(fd, dyns, PHDR(p_filesz), PHDR(p_offset)) != (int)PHDR(p_filesz))
    {
      perror ("converting RPATH to RUNPATH"), ret = 1;
      goto ret;
    }
    printf("%s: RPATH converted to RUNPATH\n", filename);
  }
#endif /* DT_RUNPATH */

  printf("%s: %s=%s\n", filename, elf_tagname(DYNSS(rpath_dyns_index, d_tag)),
         rpath);

  if (NULL == newpath)
    goto ret;

  rpathlen = strlen(rpath);

  /*
   * Calculate the maximum rpath length (will be equal to rpathlen unless
   * we have previously truncated it).
   */
  for ( i = rpathoff + rpathlen ; (i < (int)SHDR_O(sh_size)
                                   && strtab[i] == '\0') ; i++ )
    ;
  i--;

  if (i > (int)(rpathoff + rpathlen))
     rpathlen = i - rpathoff;

  if (strlen(newpath) > rpathlen)
  {
    fprintf(stderr, "new rpath '%s' too large; maximum length %i\n",
            newpath, rpathlen),
      ret = 7;
    goto ret;
  }

  strncpy(rpath, newpath, rpathlen);  // NUL-padding exploited!

  if (pwrite(fd, rpath, rpathlen, SHDR_O(sh_offset)+rpathoff) != (int)rpathlen)
  {
    perror ("writing RPATH"), ret = 1;
    goto ret;
  }
  printf("%s: new %s: %s\n", filename,
         elf_tagname(DYNSS(rpath_dyns_index, d_tag)), rpath);

ret:
  elf_close(fd);
  free(dyns);
  free(section_names);
  free(strtab);
  return ret;
}
