// SPDX-License-Identifier: GPL-2.0-or-later

/*
Taken from another list:

_Changing_ is a little tricky, but the attached program strips rpaths
from executables (I find it essential for debugging the binutils).
It's endian-dependent, if you want this for x86 you can just change
the occurrences of 'MSB' to 'LSB' and compile (I should really fix
that).

--
Geoffrey Keating <geoffk@ozemail.com.au>
*/

#include "config.h"
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <elf.h>
#if defined(HAVE_LINK_H)
#  include <link.h>
#endif /* HAVE_LINK_H */
#include <stdlib.h>
#include "protos.h"
#include <string.h>

/* Reads an ELF file, nukes all the RPATH entries. */

int
killrpath(const char *filename)
{
   int fd;
   ChrElf_Ehdr ehdr;
   int i, ret = 0;
   ChrElf_Phdr phdr;
   void *dyns = NULL;
   int dynpos;

   fd = elf_open(filename, O_RDWR, &ehdr);

   if (fd == -1)
   {
     perror ("elf_open");
     return 1;
   }

   if (0 != elf_find_dynamic_section(fd, &ehdr, &phdr))
   {
     perror("found no dynamic section"), ret = 1;
     goto ret;
   }

   dyns = malloc(PHDR(p_memsz));
   if (dyns == NULL)
     {
       perror ("allocating memory for dynamic section"), ret = 1;
       goto ret;
     }
   if (pread(fd, dyns, PHDR(p_filesz), PHDR(p_offset)) != (ssize_t)PHDR(p_filesz))
     {
       perror ("reading dynamic section"), ret = 1;
       goto ret;
     }

   dynpos = 0;
   for (i = 0; DYNSS(i, d_tag) != DT_NULL; i++)
     {
       if (is_e32) {
        ((Elf32_Dyn *)dyns)[dynpos] = ((Elf32_Dyn *)dyns)[i];

        /* DT_MIPS_RLD_MAP_REL is relative to the offset of the tag.
           Adjust it consequently.  */
        if (DYNSS(i, d_tag) == DT_MIPS_RLD_MAP_REL)
         ((Elf32_Dyn *)dyns)[dynpos].d_un.d_val =
          DO_SWAPU32(DYNSU(i, d_un.d_val) +
          (i - dynpos) * sizeof(Elf32_Dyn));
       } else {
        ((Elf64_Dyn *)dyns)[dynpos] = ((Elf64_Dyn *)dyns)[i];

        /* Ditto */
        if (DYNSS(i, d_tag) == DT_MIPS_RLD_MAP_REL)
         ((Elf64_Dyn *)dyns)[dynpos].d_un.d_val =
          DO_SWAPU64(DYNSU(i, d_un.d_val) +
          (i - dynpos) * sizeof(Elf64_Dyn));
       }
       if ( ! elf_dynpath_tag(DYNSS(i, d_tag)) )
        dynpos++;
     }
   for (; dynpos < i; dynpos++)
     {
       if (is_e32) {
        ((Elf32_Dyn *)dyns)[dynpos].d_tag = DT_NULL;
        ((Elf32_Dyn *)dyns)[dynpos].d_un.d_val = 0x0;
       } else {
        ((Elf64_Dyn *)dyns)[dynpos].d_tag = DT_NULL;
        ((Elf64_Dyn *)dyns)[dynpos].d_un.d_val = 0x0;
       }
     }

   if (pwrite(fd, dyns, PHDR(p_filesz), PHDR(p_offset)) != (int)PHDR(p_filesz))
     {
       perror ("writing dynamic section"), ret = 1;
       goto ret;
     }

ret:
   free(dyns);
   elf_close(fd);
   return ret;
}
