// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Author: Petter Reinholdtsen <pere@hungry.com>
 * date:   2001-01-20
 *
 * Alter ELF rpath information (insert, modify, remove).
 *
 * Based on source from Peeter Joot <peeterjoot@protonmail.com>
 *                  and Geoffrey Keating <geoffk@ozemail.com.au>.
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include "protos.h"

#ifdef HAVE_GETOPT_LONG
#  define GETOPT_LONG getopt_long

static const struct option long_options[] =
{
#if defined(DT_RUNPATH)
  {"convert",   no_argument,       NULL, 'c'},
#endif
  {"delete",    no_argument,       NULL, 'd'},
  {"help",      no_argument,       NULL, 'h'},
  {"keepgoing", no_argument,       NULL, 'k'},
  {"list",      no_argument,       NULL, 'l'},
  {"replace",   required_argument, NULL, 'r'},
  {"version",   no_argument,       NULL, 'v'},
  {}
};

#else /* not HAVE_GETOPT_LONG */
#  define GETOPT_LONG(argc,argv,optstr,lopts,lidx) getopt(argc,argv,optstr)
#endif /* not HAVE_GETOPT_LONG */

static void
usage(char *progname)
{
  printf("usage: %s [-k] [-l]     program...\n", progname);
  printf("       %s [-k]  -r path program...\n", progname);
  printf("       %s [-k]  -d      program...\n", progname);
#if defined(DT_RUNPATH)
  printf("       %s [-k]  -c      program...\n", progname);
#endif /* DT_RUNPATH */
  printf("\n");
  printf("  -l, --list            List the current rpath/runpath (default)\n");
  printf("  -r, --replace path    Replace current rpath/runpath setting with the path given\n");
  printf("  -d, --delete          Delete current rpath/runpath setting\n");
#if defined(DT_RUNPATH)
  printf("  -c, --convert         Convert rpath to runpath\n");
#endif /* DT_RUNPATH */
  printf("  -k, --keepgoing       Do not fail on first error\n");
  printf("  -v, --version         Display program version number\n");
  printf("  -h, --help            Show this usage information.\n");
#if !defined(DT_RUNPATH)
  printf("\n *** There is no support for runpath on this platform\n");
#endif /* DT_RUNPATH */
}

int
main(int argc, char * const argv[])
{
  bool convert = 0;     /* convert to given type */
  bool remove = 0;      /* remove or not */
  bool keep_going = 0;  /* Break on first error, or keep going? */
  const char *newpath = NULL; /* insert this path */

  if (argc < 2)
    {
      usage(argv[0]);
      return 1;
    }

  for (int opt; (opt = GETOPT_LONG(argc, argv,
#if defined(DT_RUNPATH)
    "c"
#endif
    "dhklr:v", long_options, NULL)) != -1;)
    switch (opt)
      {
      case 'l': /* This is the default action */
        newpath = NULL;
        break;
      case 'c':
        convert = true;
        break;
      case 'd':
        remove = true;
        break;
      case 'k':
        keep_going = true;
        break;
      case 'r':
        newpath = optarg;
        break;
      case 'v':
        printf("%s version %s\n", PACKAGE, VERSION);
        return 0;
      case 'h':
        usage(argv[0]);
        return 0;
      default:
        return 1;
      }

  int retval = 0;
  while (optind < argc && (!retval || keep_going))
    {
      if (remove)
        retval |= killrpath(argv[optind++]);
      else
        /* list by default, replace if path is set */
        retval |= chrpath(argv[optind++], newpath, convert);
    }

  return retval;
}
